@@ -22,8 +22,11 @@
 A tool that automatically formats Python code to conform to the PEP 8 style
 guide.
 """
+from __future__ import print_function
+
 import copy
 import os
+import re
 import sys
 import inspect
 import codecs
@@ -36,7 +39,7 @@ import token
 import tokenize
 from optparse import OptionParser
 from subprocess import Popen, PIPE
-from difflib import unified_diff
+import difflib
 import tempfile
 
 from distutils.version import StrictVersion
@@ -48,14 +51,13 @@ except ImportError:
     pep8 = None
 
 
-__version__ = '0.8.1'
+__version__ = '0.8.2'
 
 
 PEP8_BIN = 'pep8'
 CR = '\r'
 LF = '\n'
 CRLF = '\r\n'
-MAX_LINE_WIDTH = 79
 
 
 def open_with_encoding(filename, encoding, mode='r'):
@@ -64,7 +66,8 @@ def open_with_encoding(filename, encoding, mode='r'):
         # Python 3
         return open(filename, mode=mode, encoding=encoding)
     except TypeError:
-        return open(filename, mode=mode)
+        # Python 2
+        return codecs.open(filename, mode=mode, encoding=encoding)
 
 
 def detect_encoding(filename):
@@ -84,7 +87,16 @@ def detect_encoding(filename):
         except (SyntaxError, LookupError, UnicodeDecodeError):
             return 'latin-1'
     except AttributeError:
-        return 'utf-8'
+        # Python 2
+        encoding = 'utf-8'
+        try:
+            # Check for correctness of encoding
+            with open_with_encoding(filename, encoding) as input_file:
+                input_file.read()
+        except UnicodeDecodeError:
+            encoding = 'latin-1'
+
+        return encoding
 
 
 def read_from_filename(filename, readlines=False):
@@ -192,19 +204,22 @@ class FixPEP8(object):
                     completed_lines.update(modified_lines)
                 elif modified_lines == []:  # Empty list means no fix
                     if self.options.verbose >= 2:
-                        sys.stderr.write('Not fixing {f} on line {l}\n'.format(
-                            f=result['id'], l=result['line']))
+                        print(
+                            '--->  Not fixing {f} on line {l}'.format(
+                                f=result['id'], l=result['line']),
+                            file=sys.stderr)
                 else:  # We assume one-line fix when None
                     completed_lines.add(result['line'])
             else:
                 if self.options.verbose >= 3:
-                    sys.stderr.write("'%s' is not defined.\n" %
-                                     fixed_methodname)
+                    print("--->  '%s' is not defined." % fixed_methodname,
+                          file=sys.stderr)
                     info = result['info'].strip()
-                    sys.stderr.write('%s:%s:%s:%s\n' % (self.filename,
-                                                        result['line'],
-                                                        result['column'],
-                                                        info))
+                    print('--->  %s:%s:%s:%s' % (self.filename,
+                                                 result['line'],
+                                                 result['column'],
+                                                 info),
+                          file=sys.stderr)
 
     def fix(self):
         """Return a version of the source code with PEP 8 violations fixed."""
@@ -214,32 +229,44 @@ class FixPEP8(object):
                 self.options.ignore and self.options.ignore.split(','),
                 'select':
                 self.options.select and self.options.select.split(','),
+                'max_line_length':
+                self.options.max_line_length,
             }
             results = _execute_pep8(pep8_options, self.source)
         else:
             if self.options.verbose:
-                sys.stderr.write('Running in compatibility mode. Consider '
-                                 'upgrading to the latest pep8.\n')
+                print('Running in compatibility mode. Consider '
+                      'upgrading to the latest pep8.',
+                      file=sys.stderr)
             results = _spawn_pep8((['--ignore=' + self.options.ignore]
                                    if self.options.ignore else []) +
                                   (['--select=' + self.options.select]
                                    if self.options.select else []) +
+                                  (['--max-line-length={length}'.format(
+                                      length=self.options.max_line_length)]
+                                   if self.options.max_line_length else []) +
                                   [self.filename])
 
         if self.options.verbose:
-            sys.stderr.write('{n} issues to fix\n'.format(
-                n=len(results)))
+            progress = {}
+            for r in results:
+                if r['id'] not in progress:
+                    progress[r['id']] = set()
+                progress[r['id']].add(r['line'])
+            print('--->  {n} issue(s) to fix {progress}'.format(
+                n=len(results), progress=progress), file=sys.stderr)
 
-        self._fix_source(results)
+        self._fix_source(filter_results(source=''.join(self.source),
+                                        results=results))
         return ''.join(self.source)
 
     def fix_e101(self, _):
         """Reindent all lines."""
-        reindenter = Reindenter(self.source)
-        if reindenter.run():
-            original_length = len(self.source)
+        reindenter = Reindenter(self.source, self.newline)
+        modified_line_numbers = reindenter.run()
+        if modified_line_numbers:
             self.source = reindenter.fixed_lines()
-            return range(1, 1 + original_length)
+            return modified_line_numbers
         else:
             return []
 
@@ -310,7 +337,9 @@ class FixPEP8(object):
             return []
         ls, _, original = logical
         try:
-            rewrapper = Wrapper(original, hard_wrap=MAX_LINE_WIDTH)
+            rewrapper = Wrapper(
+                original, hard_wrap=self.options.max_line_length,
+                soft_wrap=self.options.max_line_length - 7)
         except (tokenize.TokenError, IndentationError):
             return []
         valid_indents = rewrapper.pep8_expected()
@@ -352,19 +381,18 @@ class FixPEP8(object):
             return []
 
     def fix_e121(self, result, logical):
-        """The 'peculiar indent' error for hanging indents."""
-        # fix by adjusting initial indent level
+        """Fix indentation to be a multiple of four."""
+        # Fix by adjusting initial indent level.
         return self._fix_reindent(result, logical)
 
     def fix_e122(self, result, logical):
-        """The 'absent indent' error for hanging indents."""
-        # fix by adding an initial indent
+        """Add absent indentation for hanging indentation."""
+        # Fix by adding an initial indent.
         return self._fix_reindent(result, logical)
 
     def fix_e123(self, result, logical):
-        """The 'loose fingernails' indentation level error for hanging
-        indents."""
-        # fix by deleting whitespace to the correct level
+        """Align closing bracket to match opening bracket."""
+        # Fix by deleting whitespace to the correct level.
         if not logical:
             return []
         logical_lines = logical[2]
@@ -374,20 +402,19 @@ class FixPEP8(object):
         fixed_line = (_get_indentation(logical_lines[0]) +
                       original_line.lstrip())
         if fixed_line == original_line:
-            # Fallback to slower method.
+            # Fall back to slower method.
             return self._fix_reindent(result, logical)
         else:
             self.source[line_index] = fixed_line
 
     def fix_e124(self, result, logical):
-        """The 'loose fingernails' indentation level error for visual
-        indents."""
-        # fix by inserting whitespace before the closing bracket
+        """Align closing bracket to match visual indentation."""
+        # Fix by inserting whitespace before the closing bracket.
         return self._fix_reindent(result, logical)
 
     def fix_e125(self, result, logical):
-        """The 'often not visually distinct' error."""
-        # fix by indenting the line in error to the next stop.
+        """Indent to distinguish line from next logical line."""
+        # Fix by indenting the line in error to the next stop.
         modified_lines = self._fix_reindent(result, logical, fix_distinct=True)
         if modified_lines:
             return modified_lines
@@ -398,7 +425,7 @@ class FixPEP8(object):
             self.source[line_index] = self.indent_word + original_line
 
     def fix_e126(self, result, logical):
-        """The 'spectacular indent' error for hanging indents."""
+        """Fix over-indented hanging indentation."""
         # fix by deleting whitespace to the left
         if not logical:
             return []
@@ -415,10 +442,10 @@ class FixPEP8(object):
             self.source[line_index] = fixed
 
     def fix_e127(self, result, logical):
-        """The 'interpretive dance' indentation error."""
+        """Fix visual indentation."""
         # Fix by inserting/deleting whitespace to the correct level.
         modified_lines = self._align_visual_indent(result, logical)
-        if modified_lines:
+        if modified_lines != []:
             return modified_lines
         else:
             # Fallback to slower method.
@@ -451,6 +478,7 @@ class FixPEP8(object):
             self.source[line_index] = fixed
 
     def fix_e201(self, result):
+        """Remove extraneous whitespace."""
         line_index = result['line'] - 1
         target = self.source[line_index]
         offset = result['column'] - 1
@@ -471,13 +499,14 @@ class FixPEP8(object):
             self.source[line_index] = fixed
 
     def fix_e224(self, result):
+        """Remove extraneous whitespace around operator."""
         target = self.source[result['line'] - 1]
         offset = result['column'] - 1
         fixed = target[:offset] + target[offset:].replace('\t', ' ')
         self.source[result['line'] - 1] = fixed
 
     def fix_e225(self, result):
-        """Fix whitespace around operator."""
+        """Fix missing whitespace around operator."""
         target = self.source[result['line'] - 1]
         offset = result['column'] - 1
         fixed = target[:offset] + ' ' + target[offset:]
@@ -499,6 +528,7 @@ class FixPEP8(object):
         self.source[line_index] = fixed
 
     def fix_e251(self, result):
+        """Remove whitespace around parameter '=' sign."""
         line_index = result['line'] - 1
         target = self.source[line_index]
 
@@ -556,20 +586,18 @@ class FixPEP8(object):
             self.source[line_index] = fixed
 
     def fix_e301(self, result):
+        """Add missing blank line."""
         cr = self.newline
         self.source[result['line'] - 1] = cr + self.source[result['line'] - 1]
 
     def fix_e302(self, result):
+        """Add missing 2 blank lines."""
         add_linenum = 2 - int(result['info'].split()[-1])
         cr = self.newline * add_linenum
         self.source[result['line'] - 1] = cr + self.source[result['line'] - 1]
 
-    def fix_e304(self, result):
-        line = result['line'] - 2
-        if not self.source[line].strip():
-            self.source[line] = ''
-
     def fix_e303(self, result):
+        """Remove extra blank lines."""
         delete_linenum = int(result['info'].split('(')[1].split(')')[0]) - 2
         delete_linenum = max(1, delete_linenum)
 
@@ -589,7 +617,14 @@ class FixPEP8(object):
 
         return modified_lines
 
+    def fix_e304(self, result):
+        """Remove blank line following function decorator."""
+        line = result['line'] - 2
+        if not self.source[line].strip():
+            self.source[line] = ''
+
     def fix_e401(self, result):
+        """Put imports on separate lines."""
         line_index = result['line'] - 1
         target = self.source[line_index]
         offset = result['column'] - 1
@@ -609,6 +644,7 @@ class FixPEP8(object):
         self.source[line_index] = fixed
 
     def fix_e501(self, result):
+        """Try to make lines fit within --max-line-length characters."""
         line_index = result['line'] - 1
         target = self.source[line_index]
 
@@ -621,7 +657,9 @@ class FixPEP8(object):
             tokens = list(tokenize.generate_tokens(sio.readline))
         except (tokenize.TokenError, IndentationError):
             multi_line_candidate = break_multi_line(
-                target, newline=self.newline, indent_word=self.indent_word)
+                target, newline=self.newline,
+                indent_word=self.indent_word,
+                max_line_length=self.options.max_line_length)
 
             if multi_line_candidate:
                 self.source[line_index] = multi_line_candidate
@@ -636,12 +674,16 @@ class FixPEP8(object):
         # over
         # my_long_function_name(x, y,
         #     z, ...)
-        candidate0 = _shorten_line(tokens, source, target, indent,
-                                   self.indent_word, newline=self.newline,
-                                   reverse=False)
-        candidate1 = _shorten_line(tokens, source, target, indent,
-                                   self.indent_word, newline=self.newline,
-                                   reverse=True)
+        candidate0 = _shorten_line(
+            tokens, source, target, indent,
+            self.indent_word, newline=self.newline,
+            max_line_length=self.options.max_line_length,
+            reverse=False)
+        candidate1 = _shorten_line(
+            tokens, source, target, indent,
+            self.indent_word, newline=self.newline,
+            max_line_length=self.options.max_line_length,
+            reverse=True)
         if candidate0 and candidate1:
             if candidate0.split(self.newline)[0].endswith('('):
                 self.source[line_index] = candidate0
@@ -662,6 +704,7 @@ class FixPEP8(object):
         self.source[line_index] = target.rstrip('\n\r \t\\') + self.newline
 
     def fix_e701(self, result):
+        """Put colon-separated compound statement on separate lines."""
         line_index = result['line'] - 1
         target = self.source[line_index]
         c = result['column']
@@ -672,7 +715,7 @@ class FixPEP8(object):
         self.source[result['line'] - 1] = fixed_source
 
     def fix_e702(self, result, logical):
-        """Fix multiple statements on one line."""
+        """Put semicolon-separated compound statement on separate lines."""
         logical_lines = logical[2]
 
         line_index = result['line'] - 1
@@ -689,7 +732,7 @@ class FixPEP8(object):
             return
 
         offset = result['column'] - 1
-        first = target[:offset].rstrip(';')
+        first = target[:offset].rstrip(';').rstrip()
         second = (_get_indentation(logical_lines[0]) +
                   target[offset:].lstrip(';').lstrip())
 
@@ -722,31 +765,31 @@ class FixPEP8(object):
         self.source[line_index] = ' '.join([left, new_center, right])
 
     def fix_e721(self, _):
+        """Switch to use isinstance()."""
         return self.refactor('idioms')
 
     def fix_w291(self, result):
+        """Remove trailing whitespace."""
         fixed_line = self.source[result['line'] - 1].rstrip()
         self.source[result['line'] - 1] = '%s%s' % (fixed_line, self.newline)
 
     def fix_w293(self, result):
+        """Remove trailing whitespace on blank line."""
         assert not self.source[result['line'] - 1].strip()
         self.source[result['line'] - 1] = self.newline
 
     def fix_w391(self, _):
-        source = copy.copy(self.source)
-        source.reverse()
+        """Remove trailing blank lines."""
         blank_count = 0
-        for line in source:
+        for line in reversed(self.source):
             line = line.rstrip()
             if line:
                 break
             else:
                 blank_count += 1
-        source = source[blank_count:]
-        source.reverse()
 
         original_length = len(self.source)
-        self.source = source
+        self.source = self.source[:original_length - blank_count]
         return range(1, 1 + original_length)
 
     def refactor(self, fixer_name, ignore=None):
@@ -778,6 +821,7 @@ class FixPEP8(object):
             return range(1, 1 + original_length)
 
     def fix_w601(self, _):
+        """Replace the {}.has_key() form with 'in'."""
         return self.refactor('has_key')
 
     def fix_w602(self, _):
@@ -786,9 +830,11 @@ class FixPEP8(object):
                              ignore='with_traceback')
 
     def fix_w603(self, _):
+        """Replace <> with !=."""
         return self.refactor('ne')
 
     def fix_w604(self, _):
+        """Replace backticks with repr()."""
         return self.refactor('repr')
 
 
@@ -829,8 +875,11 @@ def _get_indentword(source):
 
 def _get_indentation(line):
     """Return leading whitespace."""
-    non_whitespace_index = len(line) - len(line.lstrip())
-    return line[:non_whitespace_index]
+    if line.strip():
+        non_whitespace_index = len(line) - len(line.lstrip())
+        return line[:non_whitespace_index]
+    else:
+        return ''
 
 
 def _analyze_pep8result(result):
@@ -845,11 +894,10 @@ def _analyze_pep8result(result):
 
 
 def _get_difftext(old, new, filename):
-    try:
-        old = [unicode(o, detect_encoding(filename)) for o in old]
-    except NameError:
-        pass
-    diff = unified_diff(old, new, 'original/' + filename, 'fixed/' + filename)
+    diff = difflib.unified_diff(
+        old, new,
+        'original/' + filename,
+        'fixed/' + filename)
     return ''.join(diff)
 
 
@@ -876,9 +924,9 @@ def _priority_key(pep8_result):
 
 
 def _shorten_line(tokens, source, target, indentation, indent_word, newline,
-                  reverse=False):
+                  max_line_length, reverse=False):
     """Separate line at OPERATOR."""
-    max_line_width_minus_indentation = MAX_LINE_WIDTH - len(indentation)
+    max_line_length_minus_indentation = max_line_length - len(indentation)
     if reverse:
         tokens.reverse()
     for tkn in tokens:
@@ -886,12 +934,12 @@ def _shorten_line(tokens, source, target, indentation, indent_word, newline,
         if token.OP == tkn[0] and tkn[1] != '=':
             offset = tkn[2][1] + 1
             if reverse:
-                if offset > (max_line_width_minus_indentation -
+                if offset > (max_line_length_minus_indentation -
                              len(indent_word)):
                     continue
             else:
                 if (len(target.rstrip()) - offset >
-                        (max_line_width_minus_indentation -
+                        (max_line_length_minus_indentation -
                          len(indent_word))):
                     continue
             first = source[:offset - len(indentation)]
@@ -910,9 +958,9 @@ def _shorten_line(tokens, source, target, indentation, indent_word, newline,
                 continue
 
             # Don't modify if lines are not short enough
-            if len(first) > max_line_width_minus_indentation:
+            if len(first) > max_line_length_minus_indentation:
                 continue
-            if len(second) > MAX_LINE_WIDTH:  # Already includes indentation
+            if len(second) > max_line_length:  # Already includes indentation
                 continue
             # Do not begin a line with a comma
             if second.lstrip().startswith(','):
@@ -942,8 +990,7 @@ def fix_whitespace(line, offset, replacement):
 
 def _spawn_pep8(pep8_options):
     """Execute pep8 via subprocess.Popen."""
-    paths = os.environ['PATH'].split(':')
-    for path in paths:
+    for path in os.environ['PATH'].split(':'):
         if os.path.exists(os.path.join(path, PEP8_BIN)):
             cmd = ([os.path.join(path, PEP8_BIN)] +
                    pep8_options)
@@ -995,7 +1042,9 @@ class Reindenter(object):
 
     """
 
-    def __init__(self, input_text):
+    def __init__(self, input_text, newline):
+        self.newline = newline
+
         self.find_stmt = 1  # next token begins a fresh stmt?
         self.level = 0  # current indent level
 
@@ -1003,11 +1052,22 @@ class Reindenter(object):
         self.raw = input_text
         self.after = None
 
+        self.string_content_line_numbers = multiline_string_lines(
+            ''.join(self.raw))
+
         # File lines, rstripped & tab-expanded.  Dummy at start is so
         # that we can use tokenize's 1-based line numbering easily.
-        # Note that a line is all-blank iff it's "\n".
-        self.lines = [line.rstrip('\n \t').expandtabs() + '\n'
-                      for line in self.raw]
+        # Note that a line is all-blank iff it is a newline.
+        self.lines = []
+        for line_number, line in enumerate(self.raw, start=1):
+            # Do not modify if inside a multi-line string.
+            if line_number in self.string_content_line_numbers:
+                self.lines.append(line)
+            else:
+                # Only expand leading tabs.
+                self.lines.append(_get_indentation(line).expandtabs() +
+                                  line.strip() + newline)
+
         self.lines.insert(0, None)
         self.index = 1  # index into self.lines of next line
 
@@ -1018,15 +1078,20 @@ class Reindenter(object):
         self.stats = []
 
     def run(self):
+        """Fix indentation and return modified line numbers.
+
+        Line numbers are indexed at 1.
+
+        """
         tokens = tokenize.generate_tokens(self.getline)
         try:
             for t in tokens:
                 self.tokeneater(*t)
         except (tokenize.TokenError, IndentationError):
-            return False
+            return set()
         # Remove trailing empty lines.
         lines = self.lines
-        while lines and lines[-1] == '\n':
+        while lines and lines[-1] == self.newline:
             lines.pop()
         # Sentinel.
         stats = self.stats
@@ -1081,16 +1146,24 @@ class Reindenter(object):
             if diff == 0 or have == 0:
                 after.extend(lines[thisstmt:nextstmt])
             else:
-                for line in lines[thisstmt:nextstmt]:
-                    if diff > 0:
-                        if line == '\n':
+                for line_number, line in enumerate(lines[thisstmt:nextstmt],
+                                                   start=thisstmt):
+                    if line_number in self.string_content_line_numbers:
+                        after.append(line)
+                    elif diff > 0:
+                        if line == self.newline:
                             after.append(line)
                         else:
                             after.append(' ' * diff + line)
                     else:
                         remove = min(_leading_space_count(line), -diff)
                         after.append(line[remove:])
-        return self.raw != self.after
+
+        if self.raw == self.after:
+            return set()
+        else:
+            return (set(range(1, 1 + len(self.raw))) -
+                    self.string_content_line_numbers)
 
     def fixed_lines(self):
         return self.after
@@ -1167,21 +1240,18 @@ class Wrapper(object):
                 continue
             if t[0] != tokenize.ENDMARKER:
                 self.tokens.append(t)
-        self.logical_line, self.mapping = self.build_tokens_logical(
-            self.tokens
-        )
+
+        self.logical_line = self.build_tokens_logical(self.tokens)
 
     def build_tokens_logical(self, tokens):
         """Build a logical line from a list of tokens.
 
-        Returns the logical line and a list of (offset, token) tuples.  Does
+        Return the logical line and a list of (offset, token) tuples. Does
         not mute strings like the version in pep8.py.
 
         """
         # from pep8.py with minor modifications
-        mapping = []
         logical = []
-        length = 0
         previous = None
         for t in tokens:
             token_type, text = t[0:2]
@@ -1195,29 +1265,24 @@ class Wrapper(object):
                     if prev_text == ',' or (prev_text not in '{[('
                                             and text not in '}])'):
                         logical.append(' ')
-                        length += 1
                 elif end != start:  # different column
                     fill = self.lines[end_line - 1][end:start]
                     logical.append(fill)
-                    length += len(fill)
-            mapping.append((length, t))
             logical.append(text)
-            length += len(text)
             previous = t
         logical_line = ''.join(logical)
         assert logical_line.lstrip() == logical_line
         assert logical_line.rstrip() == logical_line
-        return logical_line, mapping
+        return logical_line
 
     def pep8_expected(self):
-        """Replicates logic in pep8.py, to know what level to indent things to.
+        """Replicate logic in pep8.py, to know what level to indent things to.
 
-        Returns a list of lists; each list represents valid indent levels for
+        Return a list of lists; each list represents valid indent levels for
         the line in question, relative from the initial indent.  However, the
         first entry is the indent level which was expected.
 
         """
-
         # What follows is an adjusted version of
         # pep8.py:continuation_line_indentation. All of the comments have been
         # stripped and the 'yield' statements replaced with 'pass'.
@@ -1423,7 +1488,7 @@ def refactor_with_2to3(source_text, fixer_name):
         return str(tool.refactor_string(source_text, name=''))
 
 
-def break_multi_line(source_text, newline, indent_word):
+def break_multi_line(source_text, newline, indent_word, max_line_length):
     """Break first line of multi-line code.
 
     Return None if a break is not possible.
@@ -1432,7 +1497,7 @@ def break_multi_line(source_text, newline, indent_word):
     # Handle special case only.
     if ('(' in source_text and source_text.rstrip().endswith(',')):
         index = 1 + source_text.find('(')
-        if index >= MAX_LINE_WIDTH:
+        if index >= max_line_length:
             return None
 
         # Make sure we are not in a string.
@@ -1463,13 +1528,76 @@ def check_syntax(code):
         return False
 
 
+def filter_results(source, results):
+    """Filter out spurious reports from pep8.
+
+    Currently we filter out errors about indentation in multiline strings.
+
+    """
+    string_line_numbers = multiline_string_lines(source)
+
+    for r in results:
+        if r['line'] in string_line_numbers:
+            if r['id'].lower().startswith('e1'):
+                continue
+            elif r['id'].lower() in ['e501', 'w191']:
+                continue
+
+        # Filter out incorrect E101 reports when there are no tabs.
+        # pep8 will complain about this even if the tab indentation found
+        # elsewhere is in a multi-line string.
+        if r['id'].lower() == 'e101' and '\t' not in source[r['line'] - 1]:
+            continue
+
+        yield r
+
+
+def multiline_string_lines(source):
+    """Return line numbers that are within multiline strings.
+
+    The line numbers are indexed at 1.
+
+    Docstrings are ignored.
+
+    """
+    sio = StringIO(source)
+    line_numbers = set()
+    previous_token_type = ''
+    try:
+        for t in tokenize.generate_tokens(sio.readline):
+            token_type = t[0]
+            token_string = t[1]
+            start_row = t[2][0]
+            end_row = t[3][0]
+
+            if (token_type == tokenize.STRING and
+                    starts_with_triple(token_string) and
+                    previous_token_type != tokenize.INDENT):
+                # We increment by one since we want the contents of the
+                # string.
+                line_numbers |= set(range(1 + start_row, 1 + end_row))
+
+            previous_token_type = token_type
+    except (IndentationError, tokenize.TokenError):
+        pass
+
+    return line_numbers
+
+
+def starts_with_triple(string):
+    """Return True if the string starts with triple single/double quotes."""
+    return (string.strip().startswith('"""') or
+            string.strip().startswith("'''"))
+
+
 def fix_file(filename, opts, output=sys.stdout):
     tmp_source = read_from_filename(filename)
 
     # Add missing newline (important for diff)
-    tmp_newline = find_newline(tmp_source)
-    if tmp_source == tmp_source.rstrip(tmp_newline):
-        tmp_source += tmp_newline
+    if tmp_source:
+        tmp_newline = find_newline(tmp_source)
+        if tmp_source == tmp_source.rstrip(tmp_newline):
+            tmp_source += tmp_newline
 
     fix = FixPEP8(filename, opts, contents=tmp_source)
     fixed_source = fix.fix()
@@ -1478,25 +1606,33 @@ def fix_file(filename, opts, output=sys.stdout):
     if not pep8 or opts.in_place:
         encoding = detect_encoding(filename)
 
-    # For Python3
+    interruption = None
     try:
-        tmp_source = unicode(tmp_source, detect_encoding(filename))
-    except NameError:
-        pass
+        # Keep a history to break out of cycles.
+        old_tmp_source = None
 
-    for _ in range(opts.pep8_passes):
-        if fixed_source == tmp_source:
-            break
-        tmp_source = copy.copy(fixed_source)
-        if not pep8:
-            tmp_filename = tempfile.mkstemp()[1]
-            fp = open_with_encoding(tmp_filename, encoding=encoding, mode='w')
-            fp.write(fixed_source)
-            fp.close()
-        fix = FixPEP8(tmp_filename, opts, contents=tmp_source)
-        fixed_source = fix.fix()
-        if not pep8:
-            os.remove(tmp_filename)
+        for _ in range(opts.pep8_passes):
+            if fixed_source == tmp_source:
+                break
+            if fixed_source == old_tmp_source:
+                break
+
+            old_tmp_source = tmp_source
+            tmp_source = copy.copy(fixed_source)
+
+            if not pep8:
+                tmp_filename = tempfile.mkstemp()[1]
+                fp = open_with_encoding(tmp_filename,
+                                        encoding=encoding, mode='w')
+                fp.write(fixed_source)
+                fp.close()
+            fix = FixPEP8(tmp_filename, opts, contents=tmp_source)
+            fixed_source = fix.fix()
+            if not pep8:
+                os.remove(tmp_filename)
+    except KeyboardInterrupt as exception:
+        # Allow stopping early.
+        interruption = exception
     del tmp_filename
     del tmp_source
 
@@ -1512,6 +1648,9 @@ def fix_file(filename, opts, output=sys.stdout):
     else:
         output.write(fixed_source)
 
+    if interruption:
+        raise interruption
+
 
 def parse_args(args):
     """Parse command-line options."""
@@ -1532,16 +1671,22 @@ def parse_args(args):
                       help='run recursively; must be used with --in-place or '
                            '--diff')
     parser.add_option('-p', '--pep8-passes',
-                      default=100, type='int',
+                      default=100, type=int,
                       help='maximum number of additional pep8 passes'
                            ' (default: %default)')
+    parser.add_option('--list-fixes', action='store_true',
+                      help='list codes for fixes; '
+                           'used by --ignore and --select')
     parser.add_option('--ignore', default='',
                       help='do not fix these errors/warnings (e.g. E4,W)')
     parser.add_option('--select', default='',
                       help='fix only these errors/warnings (e.g. E4,W)')
+    parser.add_option('--max-line-length', default=79, type=int,
+                      help='set maximum allowed line length '
+                           '(default: %default)')
     opts, args = parser.parse_args(args)
 
-    if not len(args):
+    if not len(args) and not opts.list_fixes:
         parser.error('incorrect number of arguments')
 
     if len(args) > 1 and not (opts.in_place or opts.diff):
@@ -1555,12 +1700,37 @@ def parse_args(args):
     if opts.in_place and opts.diff:
         parser.error('--in-place and --diff are mutually exclusive')
 
+    if opts.max_line_length < 8:
+        parser.error('--max-line-length must greater than 8')
+
     return opts, args
 
 
+def supported_fixes():
+    """Yield pep8 error codes that autopep8 fixes.
+
+    Each item we yield is a tuple of the code followed by its description.
+
+    """
+    instance = FixPEP8(filename=None, options=None, contents='')
+    for attribute in dir(instance):
+        code = re.match('fix_([ew][0-9][0-9][0-9])', attribute)
+        if code:
+            yield (code.group(1).upper(),
+                   re.sub(r'\s+', ' ',
+                          getattr(instance, attribute).__doc__))
+
+
 def main():
     """Tool main."""
     opts, args = parse_args(sys.argv[1:])
+
+    if opts.list_fixes:
+        for code, description in supported_fixes():
+            print('{code} - {description}'.format(
+                code=code, description=description))
+        return 0
+
     if opts.in_place or opts.diff:
         filenames = list(set(args))
     else:
@@ -1585,11 +1755,11 @@ def main():
                         directories.remove(d)
         else:
             if opts.verbose:
-                sys.stderr.write('[file:%s]\n' % name)
+                print('[file:%s]' % name, file=sys.stderr)
             try:
                 fix_file(name, opts, output)
-            except (UnicodeDecodeError, UnicodeEncodeError, IOError) as error:
-                sys.stderr.write(str(error) + '\n')
+            except IOError as error:
+                print(str(error), file=sys.stderr)
 
 
 if __name__ == '__main__':
