@@ -91,21 +91,75 @@ class TestUtils(unittest.TestCase):
             'foo_bar_zap_bing_bang_boom(\n    111, 111, 111, 111, 222, 222, 222, 222, 222, 222, 222, 222, 222, 333, 333,\n',
             autopep8.break_multi_line(
                 'foo_bar_zap_bing_bang_boom(111, 111, 111, 111, 222, 222, 222, 222, 222, 222, 222, 222, 222, 333, 333,\n',
-                newline='\n', indent_word='    '))
+                newline='\n', indent_word='    ',
+                max_line_length=79))
 
     def test_break_multi_line_should_not_break_too_long_line(self):
         self.assertEqual(
             None,
             autopep8.break_multi_line(
                 'foo_bar_zap_bing_bang_boom_foo_bar_zap_bing_bang_boom_foo_bar_zap_bing_bang_boom(333,\n',
-                newline='\n', indent_word='    '))
+                newline='\n', indent_word='    ',
+                max_line_length=79))
 
     def test_break_multi_line_should_not_modify_comment(self):
         self.assertEqual(
             None,
             autopep8.break_multi_line(
                 '# foo_bar_zap_bing_bang_boom(111, 111, 111, 111, 222, 222, 222, 222, 222, 222, 222, 222, 222, 333, 333,\n',
-                newline='\n', indent_word='    '))
+                newline='\n', indent_word='    ',
+                max_line_length=79))
+
+    def test_multiline_string_lines(self):
+        self.assertEqual(
+            set([2]),
+            autopep8.multiline_string_lines(
+                """
+'''
+'''
+""".lstrip()))
+
+    def test_multiline_string_lines_with_many(self):
+        self.assertEqual(
+            set([2, 7, 10, 11, 12]),
+            autopep8.multiline_string_lines(
+                """
+'''
+'''
+''''''
+''''''
+''''''
+'''
+'''
+
+'''
+
+
+'''
+""".lstrip()))
+
+    def test_multiline_string_should_not_report_single_line(self):
+        self.assertEqual(
+            set(),
+            autopep8.multiline_string_lines(
+                """
+'''abc'''
+""".lstrip()))
+
+    def test_multiline_string_should_not_report_docstrings(self):
+        self.assertEqual(
+            set([5]),
+            autopep8.multiline_string_lines(
+                """
+def foo():
+    '''Foo.
+    Bar.'''
+    hello = '''
+'''
+""".lstrip()))
+
+    def test_supported_fixes(self):
+        self.assertIn('E101', [f[0] for f in autopep8.supported_fixes()])
 
 
 class TestFixPEP8Error(unittest.TestCase):
@@ -119,12 +173,16 @@ class TestFixPEP8Error(unittest.TestCase):
     def _inner_setup(self, line, options=()):
         with open(self.tempfile[1], 'w') as temp_file:
             temp_file.write(line)
-        opts, _ = autopep8.parse_args(list(options) + [self.tempfile[1]])
+        opts, _ = autopep8.parse_args([self.tempfile[1]] + list(options))
         sio = StringIO()
+        if sys.version_info[0] < 3:
+            output = codecs.getwriter(locale.getpreferredencoding())(sio)
+        else:
+            output = sio
         autopep8.fix_file(filename=self.tempfile[1],
                           opts=opts,
-                          output=sio)
-        self.result = sio.getvalue()
+                          output=output)
+        self.result = output.getvalue()
 
     def test_e101(self):
         line = """
@@ -136,6 +194,94 @@ while True:
 while True:
     if True:
         1
+""".lstrip()
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
+    def test_e101_should_not_expand_non_indentation_tabs(self):
+        line = """
+while True:
+    if True:
+    \t1 == '\t'
+""".lstrip()
+        fixed = """
+while True:
+    if True:
+        1 == '\t'
+""".lstrip()
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
+    def test_e101_should_ignore_multiline_strings(self):
+        line = """
+x = '''
+while True:
+    if True:
+    \t1
+'''
+""".lstrip()
+        fixed = """
+x = '''
+while True:
+    if True:
+    \t1
+'''
+""".lstrip()
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
+    def test_e101_should_fix_docstrings(self):
+        line = """
+class Bar(object):
+    def foo():
+        '''
+\tdocstring
+        '''
+""".lstrip()
+        fixed = """
+class Bar(object):
+    def foo():
+        '''
+        docstring
+        '''
+""".lstrip()
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
+    def test_e101_when_pep8_mistakes_first_tab_in_string(self):
+        # pep8 will complain about this even if the tab indentation found
+        # elsewhere is in a multi-line string.
+        line = """
+x = '''
+\tHello.
+'''
+if True:
+    123
+""".lstrip()
+        fixed = """
+x = '''
+\tHello.
+'''
+if True:
+    123
+""".lstrip()
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
+    def test_e101_should_ignore_multiline_strings_complex(self):
+        line = """
+print(3 <> 4, '''
+while True:
+    if True:
+    \t1
+\t''', 4 <> 5)
+""".lstrip()
+        fixed = """
+print(3 != 4, '''
+while True:
+    if True:
+    \t1
+\t''', 4 != 5)
 """.lstrip()
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
@@ -174,6 +320,18 @@ try:
         self._inner_setup(line)
         self.assertEqual(self.result, line)
 
+    def test_e101_skip_innocuous(self):
+        # pep8 will complain about this even if the tab indentation found
+        # elsewhere is in a multi-line string. If we don't filter the innocuous
+        # report properly, the below command will take a long time.
+        p = Popen(list(AUTOPEP8_CMD_TUPLE) +
+                  ['-vvv', '--select=E101', '--diff',
+                   os.path.join(ROOT_DIR, 'test', 'e101_example.py')],
+                  stdout=PIPE, stderr=PIPE)
+        output, error = [x.decode('utf-8') for x in p.communicate()]
+        self.assertIn('Not fixing E101 on line', error)
+        self.assertEqual('', output)
+
     def test_e111_short(self):
         line = "class Dummy:\n  def __init__(self):\n    pass\n"
         fixed = "class Dummy:\n    def __init__(self):\n        pass\n"
@@ -263,6 +421,22 @@ def foo():
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
+    def test_e111_should_not_modify_string_contents(self):
+        line = """
+if True:
+ x = '''
+ 1
+ '''
+""".lstrip()
+        fixed = """
+if True:
+    x = '''
+ 1
+ '''
+""".lstrip()
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_e12_reindent(self):
         line = """
 
@@ -561,6 +735,22 @@ if True:
         self._inner_setup(line, options=['--select=E12'])
         self.assertEqual(self.result, fixed)
 
+    def test_e127_align_visual_indent(self):
+        line = """
+def draw(self):
+    color = [([0.2, 0.1, 0.3], [0.2, 0.1, 0.3], [0.2, 0.1, 0.3]),
+               ([0.9, 0.3, 0.5], [0.5, 1.0, 0.5], [0.3, 0.3, 0.9])  ][self._p._colored ]
+    self.draw_background(color)
+""".lstrip()
+        fixed = """
+def draw(self):
+    color = [([0.2, 0.1, 0.3], [0.2, 0.1, 0.3], [0.2, 0.1, 0.3]),
+             ([0.9, 0.3, 0.5], [0.5, 1.0, 0.5], [0.3, 0.3, 0.9])][self._p._colored]
+    self.draw_background(color)
+""".lstrip()
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_e12_with_backslash(self):
         line = r"""
 if True:
@@ -944,6 +1134,12 @@ print(111, 111, 111, 111, 222, 222, 222, 222, 222, 222, 222, 222, 222, 333,
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
+    def test_e501_with_shorter_length(self):
+        line = "foooooooooooooooooo('abcdefghijklmnopqrstuvwxyz')\n"
+        fixed = "foooooooooooooooooo(\n    'abcdefghijklmnopqrstuvwxyz')\n"
+        self._inner_setup(line, options=['--max-line-length=40'])
+        self.assertEqual(self.result, fixed)
+
     def test_e501_with_indent(self):
         line = """
 
@@ -974,7 +1170,6 @@ if True:
         self._inner_setup(line, options=['--select=E501'])
         self.assertEqual(self.result, fixed)
 
-
     def test_e501_alone_with_tuple(self):
         line = """
 
@@ -1182,6 +1377,42 @@ if True:
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
+    def test_e702_with_whitespace(self):
+        line = "print 1 ; print 2\n"
+        fixed = "print 1\nprint 2\n"
+        self._inner_setup(line, options=['--select=E702'])
+        self.assertEqual(self.result, fixed)
+
+    def test_e702_with_non_ascii_file(self):
+        line = """
+# -*- coding: utf-8 -*-
+# French comment with accent é
+# Un commentaire en français avec un accent é
+
+import time
+
+time.strftime('%d-%m-%Y');
+""".lstrip()
+
+        fixed = """
+# -*- coding: utf-8 -*-
+# French comment with accent é
+# Un commentaire en français avec un accent é
+
+import time
+
+time.strftime('%d-%m-%Y')
+""".lstrip()
+
+        if sys.version_info[0] < 3:
+            line = unicode(line, 'utf-8')
+            fixed = unicode(fixed, 'utf-8')
+            self._inner_setup(line.encode('utf-8'))
+            self.assertEqual(self.result, fixed.encode('utf-8'))
+        else:
+            self._inner_setup(line)
+            self.assertEqual(self.result, fixed)
+
     def test_e702_with_escaped_newline(self):
         line = '1; \\\n2\n'
         fixed = '1\n2\n'
@@ -1331,6 +1562,56 @@ class TestFixPEP8Warning(unittest.TestCase):
                           output=output)
         self.result = output.getvalue()
 
+    def test_w191_should_ignore_multiline_strings(self):
+        line = """
+print(3 <> 4, '''
+while True:
+    if True:
+    \t1
+\t''', 4 <> 5)
+if True:
+\t123
+""".lstrip()
+        fixed = """
+print(3 != 4, '''
+while True:
+    if True:
+    \t1
+\t''', 4 != 5)
+if True:
+    123
+""".lstrip()
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
+    def test_w191_should_ignore_tabs_in_strings(self):
+        line = """
+if True:
+\tx = '''
+\t\tblah
+\tif True:
+\t1
+\t'''
+if True:
+\t123
+else:
+\t32
+""".lstrip()
+        fixed = """
+if True:
+    x = '''
+\t\tblah
+\tif True:
+\t1
+\t'''
+if True:
+    123
+else:
+    32
+""".lstrip()
+        self._inner_setup(line)
+        self.assertEqual(self.result, fixed)
+
     def test_w291(self):
         line = "print 'a b '\t \n"
         fixed = "print 'a b '\n"
@@ -1437,9 +1718,20 @@ a.has_key(
         self._inner_setup(line)
         self.assertEqual(self.result, fixed)
 
+    @unittest.skipIf(sys.version_info[0] < 3, 'needs to be fixed on Python 2')
     def test_w601_with_non_ascii(self):
-        line = "## éはe\ncorrect = dict().has_key('good syntax ?')\n"
-        fixed = "## éはe\ncorrect = 'good syntax ?' in dict()\n"
+        line = """
+# -*- coding: utf-8 -*-
+## éはe
+correct = dict().has_key('good syntax ?')
+""".lstrip()
+
+        fixed = """
+# -*- coding: utf-8 -*-
+## éはe
+correct = 'good syntax ?' in dict()
+""".lstrip()
+
         if sys.version_info[0] < 3:
             line = unicode(line, 'utf-8')
             fixed = unicode(fixed, 'utf-8')
@@ -1633,7 +1925,7 @@ class TestOptions(unittest.TestCase):
             temp_file.write(line)
         p = Popen(list(AUTOPEP8_CMD_TUPLE) + [self.tempfile[1]] + options,
                   stdout=PIPE)
-        self.result = p.communicate()[0].decode('utf8')
+        self.result = p.communicate()[0].decode('utf-8')
 
     def test_diff(self):
         line = "'abc'  \n"
@@ -1641,6 +1933,10 @@ class TestOptions(unittest.TestCase):
         self._inner_setup(line, ['--diff'])
         self.assertEqual('\n'.join(self.result.split('\n')[3:]), fixed)
 
+    def test_diff_wtih_empty_file(self):
+        self._inner_setup('', ['--diff'])
+        self.assertEqual('\n'.join(self.result.split('\n')[3:]), '')
+
     def test_pep8_passes(self):
         line = "'abc'  \n"
         fixed = "'abc'\n"
@@ -1655,7 +1951,7 @@ class TestOptions(unittest.TestCase):
     def test_help(self):
         p = Popen(list(AUTOPEP8_CMD_TUPLE) + ['-h'],
                   stdout=PIPE)
-        self.assertIn('Usage:', p.communicate()[0].decode('utf8'))
+        self.assertIn('Usage:', p.communicate()[0].decode('utf-8'))
 
     def test_verbose(self):
         line = 'bad_syntax)'
@@ -1664,7 +1960,7 @@ class TestOptions(unittest.TestCase):
         f.close()
         p = Popen(list(AUTOPEP8_CMD_TUPLE) + [self.tempfile[1], '-vvv'],
                   stdout=PIPE, stderr=PIPE)
-        verbose_error = p.communicate()[1].decode('utf8')
+        verbose_error = p.communicate()[1].decode('utf-8')
         self.assertIn("'fix_e901' is not defined", verbose_error)
 
     def test_verbose_diff(self):
@@ -1675,7 +1971,7 @@ class TestOptions(unittest.TestCase):
         p = Popen(list(AUTOPEP8_CMD_TUPLE) +
                   [self.tempfile[1], '-vvv', '--diff'],
                   stdout=PIPE, stderr=PIPE)
-        verbose_error = p.communicate()[1].decode('utf8')
+        verbose_error = p.communicate()[1].decode('utf-8')
         self.assertIn("'fix_e901' is not defined", verbose_error)
 
     def test_in_place(self):
@@ -1697,9 +1993,11 @@ class TestOptions(unittest.TestCase):
         f = open(self.tempfile[1], 'w')
         f.write(line)
         f.close()
-        p = Popen(list(AUTOPEP8_CMD_TUPLE) + [self.tempfile[1], '--in-place', '--diff'],
-                  stderr=PIPE)
-        result = p.communicate()[1].decode('utf8')
+        p = Popen(
+            list(AUTOPEP8_CMD_TUPLE) + [self.tempfile[1],
+                                        '--in-place', '--diff'],
+            stderr=PIPE)
+        result = p.communicate()[1].decode('utf-8')
         self.assertIn('--in-place and --diff are mutually exclusive', result)
 
     def test_recursive(self):
@@ -1717,7 +2015,7 @@ class TestOptions(unittest.TestCase):
             p = Popen(list(AUTOPEP8_CMD_TUPLE) +
                       [temp_directory, '--recursive', '--diff'],
                       stdout=PIPE)
-            result = p.communicate()[0].decode('utf8')
+            result = p.communicate()[0].decode('utf-8')
 
             self.assertEqual(
                 "-'abc'  \n+'abc'",
@@ -1737,9 +2035,13 @@ class TestOptions(unittest.TestCase):
         f.close()
         p = Popen(list(AUTOPEP8_CMD_TUPLE) + [self.tempfile[1], '--recursive'],
                   stderr=PIPE)
-        result = p.communicate()[1].decode('utf8')
+        result = p.communicate()[1].decode('utf-8')
         self.assertIn('must be used with --in-place or --diff', result)
 
+    def test_list_fixes(self):
+        self._inner_setup('', options=['--list-fixes'])
+        self.assertIn('E101', self.result)
+
 
 class TestSpawnPEP8Process(unittest.TestCase):
 
@@ -1782,6 +2084,12 @@ class TestSpawnPEP8Process(unittest.TestCase):
         self.assertEqual(self.result, fixed)
         self.assertIn('compatibility mode', sio.getvalue())
 
+    def test_max_line_length(self):
+        line = "foooooooooooooooooo('abcdefghijklmnopqrstuvwxyz')\n"
+        fixed = "foooooooooooooooooo(\n    'abcdefghijklmnopqrstuvwxyz')\n"
+        self._inner_setup(line, options=['--max-line-length=40'])
+        self.assertEqual(self.result, fixed)
+
 
 class TestCoverage(unittest.TestCase):
 
